/*!\file udp.c \brief Stellt die UDP Funktionalitaet bereit*/
//***************************************************************************
//*            udp.c
//*
//*  Mon Jul 31 21:46:47 2006
//*  Copyright  2006  User
//*  Email
///	\ingroup network
///	\defgroup UDP Der UDP Stack fuer Mikrocontroller (udp.c)
///	\code #include "arp.h" \endcode
///	\code #include "ethernet.h" \endcode
///	\code #include "ip.h" \endcode
///	\code #include "udp.h" \endcode
///	\par Uebersicht
///		Der UDP-Stack für Mikrocontroller. Behandelt komplett den UDP-Stack
/// mit Verbindungsaufbau und Abbau.
//****************************************************************************/
//@{
#include <stdio.h>
#include "../math/math.h"
#include "arp.h"
#include "ethernet.h"
#include "ip.h"
#include "udp.h"

struct UDP_SOCKET UDP_sockettable[ MAX_UDP_CONNECTIONS ];
struct UDP_SOCKET * UDP_socket;
	
/* -----------------------------------------------------------------------------------------------------------*/
/*! Hier findet die Bearbeitung des Packetes statt welches ein UDP-Packet enthaelt. Es wir versucht die 
 * Verbindung zuzuordnen, wenn dies nicht moeglich ist wird hier abgebrochen.
 * Danach wird der Inhalt dem Socket zugeordnet und Daten in den Puffer des Benutzer kopiert.
 * \warning Zu lange UDP-Packete werden abgeschnitten.
 * \param 	packet_lenght	Gibt die Packetgroesse in Byte an die das Packet lang ist.
 * \param	ethernetbuffer	Zeiger auf das Packet.
 * \return  NONE
 */
/* -----------------------------------------------------------------------------------------------------------*/
void udp( unsigned int packet_lenght, unsigned char * ethernetbuffer )
{
	
		unsigned int i, SOCKET , Offset;
	
		struct ETH_header * ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header * IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
		struct UDP_header * UDP_packet;		// TCP_struct anlegen
		UDP_packet = ( struct UDP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];
			
		SOCKET = UDP_GetSocket( ethernetbuffer );
	
		// Socket zugeordnet ?
		if ( SOCKET == 0xffff ) return;
	
		// Socket schon mit daten belegt ?
		if ( UDP_sockettable[ SOCKET ].Socketstate == SOCKET_BUSY ) return ;
		// Hat Socket Daten ?
		if ( ( ChangeEndian16bit( UDP_packet->UDP_Datalenght ) - UDP_HEADER_LENGHT ) == 0 ) return ;
		// Größe der Daten limitieren auf Puffergröße
		i = MIN( ChangeEndian16bit( UDP_packet->UDP_Datalenght) - UDP_HEADER_LENGHT , UDP_sockettable[ SOCKET ].Bufferlenght );
		
		// Größe der Daten eintragen
		UDP_sockettable[ SOCKET ].Bufferfill = i;
		// Socket als belegt melden
		UDP_sockettable[ SOCKET ].Socketstate = SOCKET_BUSY;
		
		// Offset für UDP-Daten im Ethernetfrane berechnen
		Offset = ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 ) + UDP_HEADER_LENGHT;
		
		// Daten kompieren
		while ( i-- )
			{
				UDP_sockettable[ SOCKET ].Recivebuffer[i] = ethernetbuffer[ Offset + i ];
			}
		return;
}

/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Reistriert ein Socket in den die Daten fuer ein Verbindung gehalten werden um die ausgehenden und einghenden UDP-Packet zuzuordnen.
 * \param 	IP					Die IP-Adresse des Zielhost.
 * \param	DestinationPort		Der Zielpot des Zielhost mit den verbunden werden soll. Der Sourcport wird automatisch eingestellt. Zu beachten ist das bei einer Verbindn zu Port 67 der Sourceport auf 68 eingestellt wird.
 * \param	Bufferlenght		Groesse des Datenpuffer der vom Benutzer bereitgestellt wird. Hier werden die eingegenden UDP-Daten kopiert. Dieser Puffer sollte entsprechend der Verwendung dimensioniert sein.
 * \param	UDP_Recivebuffer	Zieger auf den Puffer der vom Benutzer bereitgestellt wird.
 * \return  Beim erfolgreichen anlegen eines Socket wird die Socketnummer zurueck gegeben. Im Fehlerfall 0xffff.
 */
/* -----------------------------------------------------------------------------------------------------------*/
unsigned int UDP_RegisterSocket( unsigned long IP, unsigned int DestinationPort, unsigned int Bufferlenght, unsigned char * UDP_Recivebuffer)
{
	unsigned int SOCKET;
	
	SOCKET = UDP_Getfreesocket();
	
	if ( SOCKET == 0xffff ) return( 0xffff );
	
	UDP_sockettable[SOCKET].Socketstate = SOCKET_READY;
	UDP_sockettable[SOCKET].DestinationPort = ChangeEndian16bit( DestinationPort );
	// wenn Zielport Bootps(67) dann Sourceport auf Bootpc(68) setzen um kommunikation mit DHCP-Server zu ermöglichen
	if ( DestinationPort == 67 )
		UDP_sockettable[SOCKET].SourcePort = ChangeEndian16bit( 68 );
	else
		UDP_sockettable[SOCKET].SourcePort =~ DestinationPort;
		
	UDP_sockettable[SOCKET].DestinationIP = IP;
	UDP_sockettable[SOCKET].Bufferfill = 0;
	UDP_sockettable[SOCKET].Bufferlenght = Bufferlenght;
	UDP_sockettable[SOCKET].Recivebuffer = UDP_Recivebuffer;
		
	if ( IP == 0xffffffff ) for( unsigned char i = 0 ; i < 6 ; i++ ) UDP_sockettable[SOCKET].MACadress[i] = 0xff;
	if ( IS_ADDR_IN_MY_SUBNET( IP, Netmask ) )
		if ( IS_BROADCAST_ADDR( IP, Netmask ) ) for( unsigned char i = 0 ; i < 6 ; i++ ) UDP_sockettable[SOCKET].MACadress[i] = 0xff;
		else GetIP2MAC( IP, &UDP_sockettable[SOCKET].MACadress );
	else GetIP2MAC( Gateway, &UDP_sockettable[SOCKET].MACadress );

	return(SOCKET);
}

/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Sendet ein UDP-Packet an einen Host.
 * \param 	SOCKET			Die Socketnummer ueber die das Packet gesendet wird.
 * param	Datalenght		Gibt die Datenlaenge der Daten in Byte an die gesendet werden sollen.
 * \param	UDP_Databuffer  Zeifer auf den Datenpuffer der gesendet werden soll.
 * \return  Bei einem Fehler beim versenden wird ungleich 0 zurueckgegeben, sonst 0.
 * \sa UDP_RegisterSocket , UDP_GetSocketState
 */
/* -----------------------------------------------------------------------------------------------------------*/
unsigned int UDP_SendPacket( unsigned int SOCKET, unsigned int Datalenght, unsigned char * UDP_Databuffer )
{
	if ( SOCKET > MAX_UDP_CONNECTIONS ) return( 0xffff );
		
	unsigned char * ethernetbuffer;
	ethernetbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + UDP_HEADER_LENGHT + Datalenght );

	struct ETH_header * ETH_packet; 		// ETH_struct anlegen
	ETH_packet = (struct ETH_header *) ethernetbuffer;
	struct IP_header * IP_packet;		// IP_struct anlegen
	IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
	IP_packet->IP_Version_Headerlen = 0x45; // Standart IPv4 und Headerlenght 20byte
	struct UDP_header * UDP_packet;		// TCP_struct anlegen
	UDP_packet = ( struct UDP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];
		
	unsigned int i;
	unsigned int offset = ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 ) + UDP_HEADER_LENGHT;

	for ( i = 0 ; i < Datalenght ; i++ ) ethernetbuffer[ offset + i ] = UDP_Databuffer[ i ];
	
	MakeIPheader( UDP_sockettable[SOCKET].DestinationIP, PROTO_UDP, UDP_HEADER_LENGHT + Datalenght , ethernetbuffer );
	MakeUDPheader( SOCKET, Datalenght, ethernetbuffer );
	MakeETHheader( UDP_sockettable[SOCKET].MACadress, ethernetbuffer );
	sendEthernetframe( ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + UDP_HEADER_LENGHT + Datalenght, ethernetbuffer);

	return(0);
}

/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Gibt den Socketstatus aus.
 * \param	SOCKET	Die Socketnummer vom abzufragen Socket.
 * \return  Den Socketstatus.
 */
/* -----------------------------------------------------------------------------------------------------------*/
unsigned int UDP_GetSocketState( unsigned int SOCKET )
	{
		return( UDP_sockettable[ SOCKET ].Socketstate );
	}

/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Gibt die Anzahl der Byte aus die sich im Puffer befinden. Diese Abfrage macht nur sinn in Verbindung mit
 * UDP_GetSocketState nachdem ein UDP-Packet empfangen worden ist und der Status fuer das auf SOCKET_BUSY steht.
 * Danach werden bis zur Freigabe durch UDP_FreeBuffer keine Daten auf den Socket mehr angenommen
 * \param	SOCKET		Die Socketnummer vom abzufragen Socket.
 * \return  Anzahl der Byte im Puffer.
 *\sa UDP_GetSocketState, UDP_FreeBuffer
 */
/* -----------------------------------------------------------------------------------------------------------*/
unsigned int UDP_GetByteInBuffer( unsigned int SOCKET )
	{
		return ( UDP_sockettable[ SOCKET ].Bufferfill );
	}

/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Gibt den UDP-Puffer wieder zum empfang frei. Danach werden wieder UDP-Daten angenommen und in den Puffer kopiert.
 * \param	SOCKET		Die Socketnummer die freigegeben werden soll.
 * \return	NONE
 */
/* -----------------------------------------------------------------------------------------------------------*/
void UDP_FreeBuffer( unsigned int SOCKET )
	{
		UDP_sockettable[ SOCKET ].Bufferfill = 0;
		UDP_sockettable[ SOCKET ].Socketstate = SOCKET_READY;
		return;
	}
	
/* -----------------------------------------------------------------------------------------------------------*/
/*!\brief Gibt das Socket wieder freu und beendet die Verbindung. Alle UDP-Packet die dann von diesen Socket empfangen werden, werden verworfen.
 * \param	SOCKET		Die Socketnummer die geschlossen werden soll.
 * \return	Es wird beim erfolgreichen schliessen der Socket 0 zurueck gegeben, sonst 0xffff.
 */
/* -----------------------------------------------------------------------------------------------------------*/
unsigned int UDP_CloseSocket( unsigned int SOCKET )
	{
		if ( SOCKET > MAX_UDP_CONNECTIONS ) return( 0xffff );
		if ( UDP_sockettable[SOCKET].Socketstate == SOCKET_NOT_USE ) return( 0xffff );
		UDP_sockettable[SOCKET].Socketstate = SOCKET_NOT_USE;
		return(0);
	}

/* -----------------------------------------------------------------------------------------------------------
wählt den richtigen Socketeintrag setzt TCP_socket auf diesen eintrag
------------------------------------------------------------------------------------------------------------*/
unsigned int UDP_GetSocket( unsigned char * ethernetbuffer )
	{
		struct ETH_header * ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header * IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
		struct UDP_header * UDP_packet;		// TCP_struct anlegen
		UDP_packet = ( struct UDP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];

		for ( unsigned int a = 0 ; a < MAX_UDP_CONNECTIONS ; a ++ ) 
			{
				if ( UDP_sockettable[a].DestinationPort == UDP_packet->UDP_SourcePort && UDP_sockettable[a].SourcePort == UDP_packet->UDP_DestinationPort && UDP_sockettable[a].DestinationIP == IP_packet->IP_SourceIP ) return(a);
				// BootP sonderstellung einräumen
				if ( UDP_sockettable[a].DestinationPort == UDP_packet->UDP_SourcePort && UDP_sockettable[a].SourcePort == UDP_packet->UDP_DestinationPort ) return(a);
			}
		return(0xffff);
	}
	
/* -----------------------------------------------------------------------------------------------------------
Bastelt den UDP-header
------------------------------------------------------------------------------------------------------------*/
unsigned int MakeUDPheader( unsigned int SOCKET, unsigned int Datalenght, unsigned char * ethernetbuffer )
{
	if ( SOCKET > MAX_UDP_CONNECTIONS ) return( 0xffff );
	
	struct ETH_header * ETH_packet; 		// ETH_struct anlegen
	ETH_packet = (struct ETH_header *) ethernetbuffer;
	struct IP_header * IP_packet;		// IP_struct anlegen
	IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
	struct UDP_header * UDP_packet;		// TCP_struct anlegen
	UDP_packet = ( struct UDP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];		
		
	UDP_packet->UDP_DestinationPort = UDP_sockettable[SOCKET].DestinationPort;
	UDP_packet->UDP_SourcePort = UDP_sockettable[SOCKET].SourcePort;
	UDP_packet->UDP_Checksum = 0;
	UDP_packet->UDP_Datalenght = ChangeEndian16bit( 8 + Datalenght );
	
	return(0);
}

/* -----------------------------------------------------------------------------------------------------------
Holt den nächsten freien Socket
------------------------------------------------------------------------------------------------------------*/
unsigned int UDP_Getfreesocket()
	{
		for ( unsigned int a = 0 ; a < MAX_UDP_CONNECTIONS ; a ++ ) 
			{
				if ( UDP_sockettable[a].Socketstate == SOCKET_NOT_USE ) return(a);
			}
		return(0xffff);
	}
//@}
